/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) <2015>, Sylvain Boily
 *
 * Sylvain Boily <sylvainboilydroid@gmail.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 *
 * Please follow coding guidelines
 * https://wiki.asterisk.org/wiki/display/AST/Coding+Guidelines
 */

/*! \file
 *
 * \brief Consul discovery module ressource
 *
 * \author\verbatim Sylvain Boily <sboily@avencall.com> \endverbatim
 *
 * This is a resource to discovery an Asterisk application via Consul
 * \ingroup applications
 */

/*! \li \ref res_discovery_consul.c uses configuration file \ref res_discovery_consul.conf
 * \addtogroup configuration_file Configuration Files
 */

/*! 
 * \page res_discovery_consul.conf res_discovery_consul.conf
 * \verbinclude res_discovery_consul.conf.sample
 */

/*** MODULEINFO
	<defaultenabled>no</defaultenabled>
	<depend>consul</depend>
	<support_level>extended</support_level>
 ***/

/*! \requirements
 *
 * asterisk - http://asterisk.org
 *
 * Build:
 *
 * make
 * make install
 * make samples
 * 
 */

#include "asterisk.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>

#include "asterisk/module.h"
#include "asterisk/config.h"
#include "asterisk/json.h"
#include "asterisk/uuid.h"
#include "asterisk/cli.h"
#include "asterisk/manager.h"
#include "asterisk/strings.h"
#include "asterisk/consul.h"
#include "asterisk/threadpool.h"


/*** DOCUMENTATION
	<configInfo name="res_discovery_consul" language="en_US">
		<synopsis>Consul client.</synopsis>
		<configFile name="res_discovery_consul.conf">
			<configObject name="general">
				<synopsis>Global configuration settings</synopsis>
				<configOption name="enabled">
					<synopsis>Enable/disable the Consul discovery module.</synopsis>
				</configOption>
			</configObject>
		</configFile>
	</configInfo>
	<manager name="DiscoverySetMaintenance" language="en_US">
		<synopsis>
			Discovery consul.
		</synopsis>
		<description>
			<para>...</para>
		</description>
	</manager>
	<managerEvent language="en_US" name="DiscoveryRegister">
		<managerEventInstance class="EVENT_FLAG_SYSTEM">
			<synopsis>Raised when are registred to consul.</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/managerEvent[@name='DiscoveryRegister']/managerEventInstance/syntax/parameter)" />
		</syntax>
		<see-also>
			<ref type="managerEvent">DiscoveryDeregister</ref>
		</see-also>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="DiscoveryDeregister">
		<managerEventInstance class="EVENT_FLAG_SYSTEM">
			<synopsis>Raised when are deregistred to consul.</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/managerEvent[@name='DiscoveryDeregister']/managerEventInstance/syntax/parameter)" />
		</syntax>
		<see-also>
			<ref type="managerEvent">DiscoveryDeregister</ref>
		</see-also>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="DiscoverySetMaintenance">
		<managerEventInstance class="EVENT_FLAG_SYSTEM">
			<synopsis>Raised when you set maintenance.</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/managerEvent[@name='DiscoveryDeregister']/managerEventInstance/syntax/parameter)" />
		</syntax>
		</managerEventInstance>
	</managerEvent>
 ***/

struct discovery_config {
	int enabled;
	char id[256];
	char eid[18];
	char name[256];
	char host[256];
	char discovery_ip[16];
	int discovery_port;
	char discovery_interface[32];
	int port;
	char tags[256];
	char token[256];
	int check;
	int check_http_port;
    int check_tls;
    char check_tls_server_name[256];
    int check_tls_skip_verify;
};

static struct discovery_config global_config = {
	.enabled = 1,
	.id = "asterisk",
	.name = "Asterisk",
	.discovery_ip = "127.0.0.1",
	.discovery_port = 5060,
	.discovery_interface = "eth0",
	.tags = "asterisk",
	.check = 0,
	.check_http_port = 8088,
    .check_tls = 0,
    .check_tls_server_name = "",
    .check_tls_skip_verify = 0
};

static const char config_file[] = "res_consul_discovery.conf";

struct ast_threadpool* discovery_thread_pool;

/*! \brief Function called to register Asterisk service into Consul */
static int consul_register(void* userdata)
{
	int success;
	struct ast_consul_service_check* checks[2] = { NULL, NULL };
	const char *tags[2] = { &global_config.tags[0], NULL };
	const char *meta[3] = { "eid", &global_config.eid[0], NULL };

    struct ast_consul_service_check httpstatus_check;
    char url_check[512];
	if (global_config.check == 1) {
		snprintf(url_check, sizeof(url_check), "%s://%s:%d/httpstatus",
                 global_config.check_tls ? "https" : "http", global_config.discovery_ip,
                 global_config.check_http_port);
		httpstatus_check.http = url_check;
		httpstatus_check.interval = 15;
        httpstatus_check.tls_server_name = ast_strlen_zero(global_config.check_tls_server_name) ? NULL
                : global_config.check_tls_server_name;
        httpstatus_check.tls_skip_verify = global_config.check_tls_skip_verify;
		checks[0]  = &httpstatus_check;
	}

	while (1) {
		success = ast_consul_service_register(
			global_config.id,
			global_config.name,
			global_config.discovery_ip,
			global_config.discovery_port,
			tags,
			meta,
			checks
		);

		if (success)
			break;

		sleep(3);
	}

	ast_log(LOG_NOTICE, "registered Consul service %s\n", global_config.id);

	manager_event(EVENT_FLAG_SYSTEM, "DiscoveryRegister", NULL);

	return success;
}

/*! \brief Function called to deregister service from Consul */
static int consul_deregister(void)
{
	if (!ast_consul_service_deregister(global_config.id)) {
		return 0;
	}

	manager_event(EVENT_FLAG_SYSTEM, "DiscoveryDeregister", NULL);

	return 1;
}

/*! \brief Function called to set maintenance state of a Consul service */
static int consul_maintenance_service(int enable)
{
	int success = ast_consul_service_set_maintenance(global_config.id, enable, "Maintenance activated by Asterisk module");

	if (success)
		manager_event(EVENT_FLAG_SYSTEM, "DiscoverySetMaintenance", "Maintenance: %s\n", enable ? "true" : "false");

	return success;
}

/*! \brief Function called to discovery ip */
static int discovery_ip_address(void)
{
	int fd;
	struct ifreq ifr;
	char host[16];

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	ifr.ifr_addr.sa_family = AF_INET;
	strncpy(ifr.ifr_name, global_config.discovery_interface, IFNAMSIZ-1);
	ioctl(fd, SIOCGIFADDR, &ifr);
	close(fd);

	sprintf(host, "%s", ast_inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr));
	ast_copy_string(global_config.discovery_ip, host, strlen(host) + 1);

	ast_debug(1,"Discovery IP: %s\n", host);

	return 0;
}

/*! \brief Function called to discovery hostname */
static int discovery_hostname(void)
{
	char hostname[1024];

	gethostname(hostname, 1024);
	ast_copy_string(global_config.name, hostname, strlen(hostname) + 1);

	ast_debug(1, "Discovery hostname: %s\n", hostname);

	return 0;
}

/*! \brief Function called to generate uuid */
static int generate_uuid_id_consul(void)
{
	const char *uuid;
	char uuid_str[256];

	uuid = ast_uuid_generate_str(uuid_str, sizeof(uuid_str));
	ast_copy_string(global_config.id, uuid, strlen(uuid) + 1);

	ast_debug(1, "Auto ID: %s\n", uuid);

	return 0;
}

/*! \brief Function called to load or reload the configuration file */
static void load_config(int reload)
{
	struct ast_config *cfg = NULL;

	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	struct ast_variable *v;

	int enabled, check, check_tls, check_tls_skip_verify;

	enabled = 1;
	check = 1;
    check_tls = 1;
    check_tls_skip_verify = 1;

	if (!(cfg = ast_config_load(config_file, config_flags)) || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "res_discovery_consul configuration file '%s' not found\n", config_file);
		return;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return;
	}

	for (v = ast_variable_browse(cfg, "general"); v; v = v->next) {
		if (!strcasecmp(v->name, "enabled")) {
			if (ast_true(v->value) == 0) {
				enabled = 0;
			}
			global_config.enabled = enabled;
		}
	}

	for (v = ast_variable_browse(cfg, "consul"); v; v = v->next) {
		if (!strcasecmp(v->name, "id")) {
			ast_copy_string(global_config.id, v->value, strlen(v->value) + 1);
		} else if (!strcasecmp(v->name, "host")) {
			ast_copy_string(global_config.host, v->value, strlen(v->value) + 1);
		} else if (!strcasecmp(v->name, "port")) {
			global_config.port = atoi(v->value);
		} else if (!strcasecmp(v->name, "tags")) {
			ast_copy_string(global_config.tags, v->value, strlen(v->value) + 1);
		} else if (!strcasecmp(v->name, "name")) {
			ast_copy_string(global_config.name, v->value, strlen(v->value) + 1);
		} else if (!strcasecmp(v->name, "discovery_ip")) {
			ast_copy_string(global_config.discovery_ip, v->value, strlen(v->value) + 1);
		} else if (!strcasecmp(v->name, "discovery_port")) {
			global_config.discovery_port = atoi(v->value);
		} else if (!strcasecmp(v->name, "discovery_interface")) {
			ast_copy_string(global_config.discovery_interface, v->value, strlen(v->value) + 1);
		} else if (!strcasecmp(v->name, "token")) {
			ast_copy_string(global_config.token, v->value, strlen(v->value) + 1);
		} else if (!strcasecmp(v->name, "check")) {
			if (ast_true(v->value) == 0) {
				check = 0;
			}
			global_config.check = check;
		} else if (!strcasecmp(v->name, "check_http_port")) {
			global_config.check_http_port =  atoi(v->value);
		} else if (!strcasecmp(v->name, "check_tls")) {
            if (ast_true(v->value) == 0) {
                check_tls = 0;
            }
            global_config.check_tls = check_tls;
        } else if (!strcasecmp(v->name, "check_tls_server_name")) {
            ast_copy_string(global_config.check_tls_server_name, v->value, strlen(v->value) + 1);
        } else if (!strcasecmp(v->name, "check_tls_skip_verify")) {
            if (ast_true(v->value) == 0) {
                check_tls_skip_verify = 0;
            }
            global_config.check_tls_skip_verify = check_tls_skip_verify;
        }
	}

	if (!strcasecmp(global_config.discovery_ip, "auto")) {
		discovery_ip_address();
	}

	if (!strcasecmp(global_config.name, "auto")) {
		discovery_hostname();
	}

	ast_eid_to_str(global_config.eid, sizeof(global_config.eid), &ast_eid_default);

	if (!strcasecmp(global_config.id, "auto")) {
		generate_uuid_id_consul();
	} else if (!strcasecmp(global_config.id, "asterisk")) {
		ast_copy_string(global_config.id, global_config.eid, strlen(global_config.eid) + 1);
	}

	ast_config_destroy(cfg);

	return;
}

/*! \brief Function called to load the resource */
static int load_res(int start)
{
	int success;

	if (start == 1) {
		struct ast_threadpool_options threadpool_opts;
		threadpool_opts.version = AST_THREADPOOL_OPTIONS_VERSION;
		threadpool_opts.idle_timeout = 15;
		threadpool_opts.auto_increment = 1;
		threadpool_opts.initial_size = 0;
		threadpool_opts.max_size = 1;
		threadpool_opts.thread_start = threadpool_opts.thread_end = NULL;

		discovery_thread_pool = ast_threadpool_create("consul-discovery", NULL, &threadpool_opts);
		if (!discovery_thread_pool) {
			ast_log(LOG_ERROR, "Failed to create Consul discovery thread pool");
			return AST_MODULE_LOAD_DECLINE;
		}

		success = ast_threadpool_push(discovery_thread_pool, consul_register, NULL) == 0;
	} else {
	    ast_threadpool_shutdown(discovery_thread_pool);
		success = consul_deregister();
	}

	return success ? AST_MODULE_LOAD_SUCCESS : AST_MODULE_LOAD_DECLINE;
}

/*! \brief Function called to exec CLI */
static char *discovery_cli_settings(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "discovery show settings";
		e->usage =
			"Usage: discovery show settings\n"
			"       Get the settings of discovery service.\n\n"
			"       Example:\n"
			"	    discovery show settings\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_cli(a->fd, "\n\nGlobal Settings:\n");
	ast_cli(a->fd, "----------------\n");
	ast_cli(a->fd, "ID service: %s\n", global_config.id);
	ast_cli(a->fd, "Name service: %s\n", global_config.name);
	ast_cli(a->fd, "Tags service: %s\n\n", global_config.tags);
	ast_cli(a->fd, "Discovery Settings:\n");
	ast_cli(a->fd, "-------------------\n");
	ast_cli(a->fd, "Discovery ip: %s\n", global_config.discovery_ip);
	ast_cli(a->fd, "Discovery port: %d\n", global_config.discovery_port);
	ast_cli(a->fd, "Discovery interface: %s\n\n", global_config.discovery_interface);
	ast_cli(a->fd, "Consul Settings:\n");
	ast_cli(a->fd, "----------------\n");
	ast_cli(a->fd, "Check: %d\n", global_config.check);
	ast_cli(a->fd, "Check http port: %d\n\n", global_config.check_http_port);
    ast_cli(a->fd, "Check TLS: %d\n", global_config.check_tls);
    ast_cli(a->fd, "Check TLSServerName: %s\n", global_config.check_tls_server_name);
    ast_cli(a->fd, "Check TLSSkipVerify: %d\n\n", global_config.check_tls_skip_verify);
	ast_cli(a->fd, "----\n");

	return NULL;
}

/*! \brief Function called to exec CLI */
static char *discovery_cli_set_maintenance(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int success;

	switch (cmd) {
	case CLI_INIT:
		e->command = "discovery set maintenance {on|off}";
		e->usage =
			"Usage: discovery set maintenance {on|off}\n"
			"       Enable/disable service in maintenance mode.\n\n"
			"       Example:\n"
			"           discovery set maintenance\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != e->args) {
		return CLI_SHOWUSAGE;
	}

	if (!strncasecmp(a->argv[e->args - 1], "on", 2)) {
		success = ast_consul_service_set_maintenance(global_config.id, 1, NULL);
		ast_cli(a->fd, "Maintenance mode for service %s is set\n", global_config.id);
	} else if (!strncasecmp(a->argv[e->args - 1], "off", 3)) {
		success = ast_consul_service_set_maintenance(global_config.id, 0, NULL);
		ast_cli(a->fd, "Maintenance mode for service %s is unset\n", global_config.id);
	}

	if (!success) {
		ast_log(LOG_NOTICE, "failed to set maintenance state of service");
	}
 
	return NULL;
}

static int manager_maintenance(struct mansession *s, const struct message *m)
{
	int success;
	const char *enable = astman_get_header(m, "Enable");

	if (ast_strlen_zero(enable)) {
			astman_send_error(s, m, "No action to enable or disable specified");
			return 0;
	}

	success = consul_maintenance_service(strcmp(enable, "true") == 0);

	return !success;
}

/*! \brief Function called to define CLI */
static struct ast_cli_entry cli_discovery[] = {
	AST_CLI_DEFINE(discovery_cli_settings, "Show discovery settings"),
	AST_CLI_DEFINE(discovery_cli_set_maintenance, "Set discovery service in maintenance mode")
};

static int reload_module(void)
{
	load_config(1);
	return 0;
}

static int unload_module(void)
{
	load_res(0);
	ast_cli_unregister_multiple(cli_discovery, ARRAY_LEN(cli_discovery));
	ast_manager_unregister("DiscoverySetMaintenance");
	return 0;
}

static int load_module(void)
{
	if (!ast_module_check("res_consul.so")) {
		if (ast_load_resource("res_consul.so") != AST_MODULE_LOAD_SUCCESS) {
			ast_log(LOG_ERROR, "Cannot load res_consul, so res_discovery_consul cannot be loaded\n");
			return AST_MODULE_LOAD_DECLINE;
		}
	}

	load_config(0);

	if (global_config.enabled == 0) {
		ast_log(LOG_NOTICE, "This module is disabled\n");
		return AST_MODULE_LOAD_DECLINE;
	}

	if (load_res(1)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_cli_register_multiple(cli_discovery, ARRAY_LEN(cli_discovery));
	ast_manager_register_xml("DiscoverySetMaintenance", EVENT_FLAG_SYSTEM, manager_maintenance);
	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "Asterisk Service Registration in Consul",
	.support_level = AST_MODULE_SUPPORT_EXTENDED,
	.load = load_module,
	.unload = unload_module,
	.reload = reload_module,
	.load_pri = AST_MODPRI_APP_DEPEND,
	.requires = "res_consul",
);
